#!/usr/bin/perl -w
#
###################################################################
#                                                                 #
#   PERL- Script                                                  #
#                                                                 #
#          Aufbereiten von Daten aus Tabellendateien              #
#          zum Einlesen in Kivitendo                              #
#                                                                 #
#                                                                 #
#          Lizenz: GPL-V3                                         #
#                     en: http://www.gnu.org/licenses/gpl.html    #
#                     de: http://www.gnu.de/documents/gpl.de.html #
#                                                                 #
#          Keine Garantie! Kein Support!                          #
#          Verwendung auf eigene Gefahr!                          #
#                                                                 #
#                                                                 #
#                            ######################################
#                            #  v0.8.3, SNR 08.2013               #
#                            #  http://www.sebastian-reinhardt.de #
#                            #  snr@arcor.de                      #
###################################################################
#
# ToDo: - weitere Kommentare einfuegen
#	- Standardwerte/ Table_args, Hersteller in config- Datei speichern(?)
#       - Importformate auf "csv" erweitern
#
use strict;
use POSIX;
use Spreadsheet::ParseExcel;
use Spreadsheet::XLSX;
use ODF::lpOD;
use URI::Escape;
use Encode::Guess;
use Encode::Encoder qw(encoder);
use Time::Local;
use Text::Iconv;
use List::MoreUtils qw(uniq);
#use PDF::API3;
use utf8;
use 5.010;
#
# output settings for console
setlocale(LC_ALL, "de_DE"); # "de"  has comma as decimal point 
binmode(STDOUT, ":utf8");
#
print "\n\n######################################################\n";
print "#                                                    #\n";
print "#                                                    #\n";
print "#      Aufbereiten von Daten aus Tabellendateien     #\n";
print "#      zum Einlesen in Kivitendo                     #\n";
print "#                                                    #\n";
print "#                                                    #\n";
print "#                                #####################\n";
print "#                                #                   #\n";
print "#                                #  by SNR, 08.2013  #\n";
print "#                                #  v0.8.3, GPL-V3   #\n";
print "#                                #                   #\n";
print "######################################################\n";
print "#                                                    #\n";
#
my %tableargs;				# hash containing rownames
my %tabletrans;				# hash containing translation of rownames
my %suppliers;				# hash containing names of suppliers/manufactorers
my %suppabbr;				# hash containing abbreviations of suppliers/manufactorers
my %unit_trans;				# hash containing translations for unit names like "Stück" => "Stck"
my %input_args;				# hash containing input arguments (e.g. input file, clone price, max. line numbers...)
my %charset;				# hash containing charset information of plain text files for inputfiles
#
#############################################################################################
#                                                                                           #
#        fill hashes                                                                        #
#                                                                                           #
#############################################################################################
#
my %input_files_hash;
%input_files_hash = ( 	'suppl_file'		=> 'config/suppliers_list.txt',
			'suppabbr_list'		=> 'config/suppabbr_list.txt',
			'global_discount'	=> 'config/glob_discount.txt',
			'unit_trans'		=> 'config/unit_trans.txt',
			'tableargs'		=> 'config/tableargs_hash.txt',
			'tabletrans'		=> 'config/tabletrans_hash.txt',
			'charset'		=> 'config/file_charset.txt'
		);
#
# initial values for some parameters
#
$input_args{'separator'} = ';' ;
$input_args{'glob_discount'} = 'config/glob_discount.txt';
$input_args{'#maxlines'} = 50000;
$input_args{'output_filenum'} = 0;
$input_args{'input_format'} ='';
#
# reading values from config files
#
my $input;
#
# list of manufactorers/ suppliers (number same as model_X/make_X!)
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'suppl_file'}) or die "Cannot open file:".$input_files_hash{'suppl_file'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $suppliers{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
	}
close($input);
#
# list of manufactorer/ supplier abbreviations (number same as model_X/make_X!)
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'suppabbr_list'}) or die "Cannot open file:".$input_files_hash{'suppabbr_list'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $suppabbr{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
	}
close($input);
#
# list of unit translationstring
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'unit_trans'}) or die "Cannot open file:".$input_files_hash{'unit_trans'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $unit_trans{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
		}
close($input);
#
# list of rownames for Kivitendo
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'tableargs'}) or die "Cannot open file:".$input_files_hash{'tableargs'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $tableargs{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
		}
close($input);
#
# list of "human readable" rownames for Kivitendo
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'tabletrans'}) or die "Cannot open file:".$input_files_hash{'tabletrans'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $tabletrans{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
		}
close($input);
#
# list of character set information of input files
#
open($input,'<:encoding(UTF-8)',$input_files_hash{'charset'}) or die "Cannot open file:".$input_files_hash{'charset'}."\n";
while (<$input>)
	{
		if( $_ !~ /#/ )
			{
				 chomp($_);
				 $charset{(split(/;;/,$_))[0]} = (split(/;;/,$_))[1];
			}
		}
close($input);
#
# parsing input arguments
#
my $inc_var=0;
while ( $inc_var le $#ARGV )
	{
		$ARGV[$inc_var]=$ARGV[$inc_var];
		#
		if( $ARGV[$inc_var] ~~ /-h/ )
			{
				&usage($ARGV[$inc_var]);
			}
		elsif( $ARGV[$inc_var] ~~ /-l/ )
			{
				&list_suppliers(\%suppliers,\%suppabbr);
			}
		elsif( $ARGV[$inc_var] ~~ /-i/)
			{
				$input_args{'input_file'} = (split(/=/,$ARGV[$inc_var]))[1];
			}
		elsif( $ARGV[$inc_var] ~~ /-o/)
			{
				$input_args{'output_file'} = (split(/=/,$ARGV[$inc_var]))[1];
			}
		elsif( $ARGV[$inc_var] ~~ /-maxlines/)
			{
				$input_args{'#maxlines'} = (split(/=/,$ARGV[$inc_var]))[1];
			}
		elsif( $ARGV[$inc_var] ~~ /-separator/)
			{
				$input_args{'separator'} = (split(/=/,$ARGV[$inc_var]))[1];
			}
		elsif( $ARGV[$inc_var] ~~ /-pnprefix/)
			{
				#
				$input_args{'pnprefix_add'} = (split(/=/,$ARGV[$inc_var]))[1];
				given( $input_args{'pnprefix_add'} )
				 	{
						when( /^\d+\z/ )
							{ 
								#
								# convert number in suppliers abbreviation
								$input_args{'pnprefix_add'} = $suppabbr{$input_args{'pnprefix_add'}};
							}
						
					}
			}
		elsif( $ARGV[$inc_var] ~~ /-desprefix/)
			{
				#
				$input_args{'desprefix_add'} = (split(/=/,$ARGV[$inc_var]))[1];
				given( $input_args{'desprefix_add'} )
				 	{
						when( /^\d+\z/ )
							{ 
								#
								# convert number in suppliers abbreviation
								$input_args{'desprefix_add'} = $suppabbr{$input_args{'desprefix_add'}};
							}
						
					}
			}
		elsif( $ARGV[$inc_var] ~~ /-cloneprice/)
			{
				$input_args{'clone_price'} = (split(/=/,$ARGV[$inc_var]))[1];
			}
		elsif( $ARGV[$inc_var] ~~ /-make_X/)
			{
				$input_args{'make_X'} = (split(/=/,$ARGV[$inc_var]))[1];
				given( $input_args{'make_X'} )
				 	{
						when( /^\d+\z/ )
							{ 
								#
								# convert number in suppliers abbreviation
								#
								$input_args{'make_X'} = $suppabbr{$input_args{'make_X'}};
							}
						
					}
			}
		elsif( $ARGV[$inc_var] ~~ /-model_X/)
			{
				$input_args{'model_X'} = (split(/=/,$ARGV[$inc_var]))[1];
				given( $input_args{'model_X'} )
				 	{
						when( /^\d+\z/ )
							{ 
								#
								# convert number in suppliers abbreviation
								$input_args{'model_X'} = $suppabbr{$input_args{'model_X'}};
							}
						
					}
			}
		elsif( $ARGV[$inc_var] ~~ /-datanorm/ )
			{
				if( (split(/=/,$ARGV[$inc_var]))[1] ~~ /5/ )
					{
						$input_args{'input_format'} = 'datanorm_50';
					}
				else
					{
						$input_args{'input_format'} = 'error_format';
					}
			}
		elsif( $ARGV[$inc_var] ~~ /-allsuppl/ )
			{
				#
				# set all supplier relevant options to given supplier
				#
				my $allsuppl=(split(/=/,$ARGV[$inc_var]))[1];
					given( $allsuppl )
						{
							when( /^\d+\z/ )
								{
									#
									# convert number in suppliers abbreviation
									$input_args{'pnprefix_add'} = $suppabbr{$allsuppl};
									#
									# convert number in suppliers abbreviation
									$input_args{'desprefix_add'} = $suppliers{$allsuppl};
									#
									# convert number in suppliers abbreviation
									#
									$input_args{'make_X'} = $allsuppl;
									#
									# convert number in suppliers abbreviation
									#
									$input_args{'model_X'} = $allsuppl;
									#
									# covert number in companies purchase price
									#
									$input_args{'lastcost_X'} = $allsuppl;
									#
									# covert number in purchase price
									#
									$input_args{'lastcost'} = $allsuppl;
									#
									# set cloneprice
									#
									$input_args{'clone_price'} = 1;
								}
						
					}
			}
		else
			{
				print "#                                                    #\n";
				print "#    Fehler: Fehlende oder unbekannte Option!        #\n";
				print "#            Bitte die Eingaben prüfen!              #\n";
				print "#                        ==> ABBRUCH!                #\n";
				&usage;
				print "#                                                    #\n";
				print "######################################################\n";
				exit 2;
			}
		$inc_var++;
	}
#
# no input given
# (error output)
#
if ( !defined($input_args{'input_file'}) )
	{
		print "#                                                    #\n";
		print "#   Fehler: Es muss als Minimum die Quelldatei       #\n";
		print "#           mit den Daten angegeben werden!          #\n";
		print "#                            ==> ABBRUCH!            #\n";
		&usage;
		print "#                                                    #\n";
		print "######################################################\n";
		exit 3
										
	}
#
# no output given
#
if ( !defined($input_args{'output_file'}) )
	{
		print "#                                                    #\n";
		print "#  Achtung: Es wurde keine Zieldatei angegeben!      #\n";
		print "#           Setzte Zieldateinamen automatisch!       #\n";
		print "#                                                    #\n";
		$input_args{'output_file'}='/tmp/kivitendo_waren_import.csv';
	}
#
print "#                                                    #\n";
print "#                     Quelldatei: ",$input_args{'input_file'},"\n";
print "#                  Zieldatei(en): ",$input_args{'output_file'},"\n";
print "#         max. Anzahl der Zeilen: ",$input_args{'#maxlines'},"\n";
print "#            Spaltentrennzeichen: ",$input_args{'separator'},"\n";
#
if ( defined($input_args{'pnprefix_add'}) )
	{
		print "#       Prefix zur Artikelnummer: ",$input_args{'pnprefix_add'},"\n";
	}
if ( defined($input_args{'desprefix_add'}) )
	{
		print "# Prefix zur Artikelbeschreibung: ",$input_args{'desprefix_add'},"\n";
	}
if ( defined($input_args{'model_X'}) )
	{
		print "# Erzeugen der Spalte Lieferantenartikelnummer\n";
		print "#      für folgenden Lieferanten: ",$suppliers{$input_args{'model_X'}},"\n";
	}
if ( defined($input_args{'make_X'}) )
	{
		print "#  Erzeugen der Spalte Lieferant: ",$suppliers{$input_args{'make_X'}},"\n";
	}
if ( $input_args{'clone_price'} ne 0 )
	{
		print "#                Preise \"clonen\": ",$input_args{'clone_price'},"\n";
	}
print "#                                                    #\n";
if ( $input_args{'#maxlines'} eq 0 )
	{
		print "#                                                    #\n";
		print "#     ACHTUNG: Die Zieldaten werden nicht in mehrere #\n";
		print "#              Dateien aufgeteilt! Um dies zu ändern #\n";
		print "#              eine Anzahl undgleich \"0\" angeben!    #\n";
		print "#                                                    #\n";
	}
#
# testing of input format
#
print "#                                                    #\n";
print "#       Erkennen des Quelldateiformates:";
#
if( ( $input_args{'input_format'} !~ /datanorm/ ) && ( (`file $input_args{'input_file'}` ~~ /Microsoft Excel/) || (`file $input_args{'input_file'}` ~~ /Composite Document File/) ) && `file $input_args{'input_file'}` !~ /Microsoft Excel 2007+/ )
	{
		print "  MS Excel format\n";
		$input_args{'input_format'}='xls';
	}
elsif( ( $input_args{'input_format'} !~ /datanorm/ ) && `file $input_args{'input_file'}` ~~ /Microsoft Excel 2007+/ )
	{
		print "  MS Excel 2007+/ xlsx format\n";
		$input_args{'input_format'}='xlsx';
	}
elsif( ( $input_args{'input_format'} !~ /datanorm/ ) && `file $input_args{'input_file'}` ~~ /OpenDocument Spreadsheet/ )
	{
		print "  OpenDocument format\n";
		$input_args{'input_format'}='ods';
	}
elsif( ( $input_args{'input_format'} !~ /datanorm/ ) && `file $input_args{'input_file'}` ~~ /text/ || `file $input_args{'input_file'}` ~~ /ascii/ )
	{
		print "  text/ ascii format\n";
		$input_args{'input_format'}='ascii';
	}
elsif( $input_args{'input_format'} ~~ /datanorm/ )
	{
		print "  text/ datanorm 5.0 format\n";
	}
elsif( lc(`file $input_args{'input_file'}`) ~~ /pdf document/ )
	{
		print "  PDF\n";
		$input_args{'input_format'} = 'pdf';
	}
else
	{
		$input_args{'input_format'}='false';
		print "#                                                    #\n";
		print "# Fehler: Kann Datei nicht verarbeiten!              #\n";
		print "#         Falsches Format?      ==> ABBRUCH!         #\n";
		print "#                                                    #\n";
		print "######################################################\n";
		exit 1;
	}
#
# parsing of input files
#
if ( ( $input_args{'input_format'} !~ /datanorm/ ) && ( $input_args{'input_format'} ~~ /ods/ ) || ( $input_args{'input_format'} ~~ /xls/ ) || ( $input_args{'input_format'} ~~ /xlsx/ ) )
	{
		#
		# parsing  spreadsheet
		#
		&parsespreadsheet(\%input_args,\%tableargs,\%tabletrans,\%suppliers,\%suppabbr,\%unit_trans);
		#
	}
elsif ( ( $input_args{'input_format'} !~ /datanorm/ ) && ( $input_args{'input_format'} ~~ /ascii/ ) )
	{
		#
		# parsing ascii files with fixed column size
		#
		&ascii(\%input_args,\%tableargs,\%tabletrans,\%suppliers,\%suppabbr,\%unit_trans);
		#
	}
elsif ( $input_args{'input_format'} ~~ /datanorm/ )
	{
		#
		# parsing ascii files with fixed column size
		#
		&datanorm_50(\%input_args,\%tableargs,\%tabletrans,\%suppliers,\%suppabbr,\%unit_trans,\%charset);
		#
	}	
elsif ( $input_args{'input_format'} ~~ /pdf/ )
	{
		#
		# parsing ascii files with fixed column size
		#
		&pdf(\%input_args,\%tableargs,\%tabletrans,\%suppliers,\%suppabbr,\%unit_trans,\%charset);
		#
	}
#
#
print "#                                                    #\n";
print "#       Umwandlung erfolgreich beendet!              #\n";
print "#                                          Danke!    #\n";
print "#                                                    #\n";
print "######################################################\n";
#
#
#	
exit 0;
#
#############################################################################################
#############################################################################################
##                                                                                         ##
##                                     Subroutines                                         ##
##                                                                                         ##
#############################################################################################
#############################################################################################
#
sub usage
	{
		my $tmp_var = shift;
		print "#                                                    #\n";
		print "#   Verwendung:                                      #\n";
		print "#    bin/list_split.pl -i=/Pfad_zur_Quelldatei/ -... #\n";
		print "#                                                    #\n";
		print "#   Optionen:                                        #\n";
		print "#    -i=/Pfad_zur_Quelldatei/Quelldatei.fmt          #\n";
		print "#    Quelldatei (MS Excel-/OpenDocument- Spreadsheet;#\n";
		print "#                relativ zum aktuellen Pfad)         #\n";
		print "#    -o=/Pfad_zur_Zieldatei/Zieldatei.csv            #\n";
		print "#    Zieldatei (CSV; wird automatisch weitergezählt; #\n";
		print "#               relativ zum aktuellen Pfad)          #\n";
		print "#    -maxlines=[int]                                 #\n";
		print "#      maximale Zeilenanzahl in der Zieldatei,       #\n";
		print "#      Vorgabe: 50000                                #\n";
		print "#      Hinweis: -maxlines=0 => eine unbegrenzte      #\n";
		print "#               Anzahl Zeilen wird geschrieben       #\n";
		print "#    -separator=[char]                               #\n";
		print "#      Spaltentrennzeichen in Zieldatei, Vorgabe:;   #\n";
		print "#    -pnprefix=[int/char]                            #\n";
		print "#      Prefix zur Artikelnummer, Möglichkeit um      #\n";
		print "#      doppelte Artikelnummern bei verschiedenen     #\n";
		print "#      Quelllisten zu vermeiden                      #\n";
		print "#    -desprefix=[int/char]                           #\n";
		print "#      Prefix zur Artikelbeschreibung hinzufügen     #\n";
		print "#    -make_X=[int/char]                              #\n";
		print "#      Spalte mit Hersteller/ Lieferanten einfügen   #\n";
		print "#    -model_X=[int/char]                             #\n";
		print "#      Spalte mit Hersteller-/ Lieferantenartikel-   #\n";
		print "#      nummer einfügen                               #\n";
		print "#    -cloneprice=[0/1]                               #\n";
		print "#      Schalter zum Erzeugen einer \"neuen\" Spalte    #\n";
		print "#      mit Preisen zu erzeugen, wenn der Listenpreis #\n";
		print "#      gleich dem Verkaufspreis ist                  #\n";
		print "#    -datanorm=[int]                                 #\n";
		print "#      Quelldateien liegen im Datanorm Format vor,   #\n";
		print "#      Es muss die Version angegeben werden!         #\n";
		print "#      derzeit wird nur die Version 5.0 unterstützt  #\n";
		print "#                                                    #\n";
		print "#    Vereinfachungen:                                #\n";
		print "#     -allsuppl=[int]                                #\n";
		print "#      entspricht folgenden Funktionen:              #\n";
		print "#              -pnprefix=[int]                       #\n";
		print "#              -desprefix=[int]                      #\n";
		print "#              -make_X=[int]                         #\n";
		print "#              -model_X=[int]                        #\n";
		print "#              -cloneprice=1                         #\n";
		print "#      Achtung: Nur mit der Nummer des Lieferanten   #\n";   
		print "#               aus Kivitendo/ den Konfigurations-   #\n";
		print "#               dateien verwendbar!                  #\n";
		print "#                                                    #\n";
		print "#    -l                                              #\n";
		print "#      Liste mit im Script hinterlegten Lieferanten  #\n";
		print "#      und diese Hilfe anzeigen                      #\n";
		print "#    -h                                              #\n";
		print "#      diese Hilfe anzeigen                          #\n";
		print "#                                                    #\n";
		print "#   HINWEIS:                                         #\n";
		print "#    Bei Schaltern/ Optionen, die mit [int/char]     #\n";
		print "#    angegeben sind, kann man die vordefinierten     #\n";
		print "#    Werte für die Hersteller/ Lieferanten mit den   #\n";
		print "#    laufenden Nummern aufrufen (sofern diese in den #\n";
		print "#    Hashes im Script hinterlegt sind) oder direkt   #\n";
		print "#    als String frei eingeben!                       #\n";
		print "#                                                    #\n";
		if( defined($tmp_var) )
			{
				if ( $tmp_var ~~ /-h/ )
					{
						print "#                                                    #\n";
						print "######################################################\n";
						exit 0
					}
			}
	}
#
#############################################################################################
#                                                                                           #
#                parse spreadsheets                                                         #
#                                                                                           #
#############################################################################################
#
sub parsespreadsheet
	{
		my %input_hash = %{(shift)};
		my %tableargs_hash = %{(shift)};
		my %tabletrans_hash = %{(shift)};
		my %suppliers_list = %{(shift)};
		my %suppabbr_list = %{(shift)};
		my %unit_trans = %{(shift)};
		my %output_hash;
		my %error_output_hash;				# hash containing errors (errornous output lines, to be written in separate file)
		my %head_hash;
		my %head_def_hash;
		my %head_trans_hash;
		my %head_col_pos;
		my %discount_hash;
		my $out_putadd;
		my $def_file;
		#
		my $skip_row = 0;
		my @head;
		my %head_hash_new;
		my %head_hash_old;
		my %head_hash_out;
		my %error_out_hash;
		#
		# parsing of input files
		#
		if ( -r 'config/'.$input_hash{'make_X'}.'_discount.txt' )
			{
				print "#                                                    #\n";
				print "#   verwende Rabattcodedatei: \"config/".$input_hash{'make_X'}."_discount.txt\"\n";
				print "#                                                    #\n";
				open($def_file,'<:encoding(UTF-8)','config/'.$input_hash{'make_X'}.'_discount.txt') or die "Could not open $def_file: $!";
				while( my $line = <$def_file> )
					{
						# read discount value
						$discount_hash{(split(/;;/,$line))[0]} = del_spaces((split(/;;/,$line))[1]);
					}
				close $def_file;
			}
		elsif ( -r $input_hash{'glob_discount'} )
			{
				print "#                                                    #\n";
				print "#   verwende Rabattcodedatei: \"".$input_hash{'glob_discount'}."\"\n";
				print "#                                                    #\n";
				open($def_file,'<:encoding(UTF-8)',$input_hash{'glob_discount'}) or die "Could not open $def_file: $!";
				while( my $line = <$def_file> )
					{
						# read discount value
						if( (split(/;;/,$line))[0] eq $input_hash{'make_X'} )
							{
								$discount_hash{'global'} = del_spaces((split(/;;/,$line))[1]);
								print "#                                                    #\n";
								print "#   anzuwendender Rabatt (global): ".$discount_hash{'global'}."\%\n";
								print "#                                                    #\n";
							}
					}
				close $def_file;
			}
		#
		my $worksheet;
		my @worksheets;
		my $workbook;
		#
		if ( $input_hash{'input_format'} eq 'xls' )
			{	
    				my $parser = Spreadsheet::ParseExcel->new();
    				$workbook = $parser->Parse($input_hash{'input_file'});
				@worksheets = $workbook->worksheets();
			}
		elsif ( $input_hash{'input_format'} eq 'xlsx' )
			{
				print "#                                                    #\n";
				print "#   Öffnen und Einlesen der Datei....                #\n";
				print "#                                                    #\n";
				$workbook = Spreadsheet::XLSX -> new($input_hash{'input_file'},);
				@worksheets = @{$workbook -> {Worksheet}};
			}
		elsif ( $input_hash{'input_format'} eq 'ods' )
			{
				$workbook = odf_document->get($input_hash{'input_file'});
				@worksheets = $workbook->get_body->get_tables;
			}
		#
		$input_hash{'output_filenum'} = 0;
		my $worksheet_i = 0;
		#     
    		foreach (@worksheets)
    			{
				if ( $worksheet_i gt 0 )
					{
						#
						# start output into file
						# (one per worksheet)
						#
						$input_hash{'output_filenum'}++;
						$out_putadd = newfile(\%head_hash_out,\%input_hash,$out_putadd);
					}
				#
				my $col_min;
				my $col_max;
				my $row_min;
				my $row_max;
				$worksheet_i++;
				print "#                                                    #\n";
				print "#       Tabellenblatt: ",$worksheet_i,"                             #\n";
				print "#                                                    #\n";
				$worksheet = $_;
				if ( $input_hash{'input_format'} eq 'xls' )
					{	
						( $row_min, $row_max ) = $worksheet->row_range();
	       					( $col_min, $col_max ) = $worksheet->col_range();
					}
				elsif ( $input_hash{'input_format'} eq 'ods' )
					{
						($row_max, $col_max) = $worksheet->get_size;
						$row_max--;
						$col_max--;
						$col_min=0;
						$row_min=0;
					}
				elsif ( $input_hash{'input_format'} eq 'xlsx' )
					{
						$row_max = $worksheet -> {MaxRow};
						$row_min = $worksheet -> {MinRow};
						$col_max = $worksheet -> {MaxCol};
						$col_min = $worksheet -> {MinCol};
						
					}
				#
				if ( $col_max > 100 )
					{
						print "#                                                    #\n";
						print "#       Fehler: Es wurden ",$col_max," Tabellenspalten\n";
						print "#               in der Datei gefunden!               #\n";
						print "#               Gibt es leere, bearbeitete Spalten   #\n";
						print "#               in der Datei?                        #\n";
						print "#       Bearbeiten Sie die Datei und versuchen Sie   #\n";
						print "#       es erneut!                                   #\n";
						print "#                                ==> ABBRUCH!        #\n";
						print "#                                                    #\n";
						print "######################################################\n";
						exit 1;
					}
				#
				# find table head line
				#
				my $tmp_skiprow = 0;
				while ($tmp_skiprow eq 0 )
					{
						for my $col ( $col_min .. $col_max )
							{
								if ( $input_hash{'input_format'} eq 'xls' )
									{
										my $cell = $worksheet->get_cell( $skip_row, $col );
										#
										# read unformatted value of cell, if cell is not empty
										#
										$head_hash_old{$col} = '';
										#
										if ( defined(uri_unescape( Encode::encode( "utf8",$cell->{Val} ))) )
											{
												$head_hash_old{$col} = uri_unescape( Encode::encode( "utf8",$cell->unformatted() ));
											}
										#
									}
								elsif ( $input_hash{'input_format'} eq 'xlsx' )
									{
										#
										my $cell = $worksheet-> {Cells} [$skip_row] [$col];
										#
										# read unformatted value of cell, if cell is not empty
										#
										$head_hash_old{$col} = '';
										#
										if ( defined( $cell ) )
											{
												$head_hash_old{$col} = uri_unescape( Encode::encode( "utf8",$cell->unformatted() ));
											}
									}
								elsif ( $input_hash{'input_format'} eq 'ods' )
									{
										#
										my $cell = $worksheet->get_cell( $skip_row, $col )->get_text;
										#
										# read unformatted value of cell, if cell is not empty
										#
										$head_hash_old{$col} = '';
										#
										if ( defined( $cell ) )
											{
												$head_hash_old{$col} = uri_unescape( $cell );
											}
									}
							}
						#
						if ( ( (grep lc($head_hash_old{$_}) ~~ /art/ , sort{$a<=>$b}(keys %head_hash_old)) gt 0 ) || ( (grep lc($head_hash_old{$_}) ~~ /nummer/ , sort{$a<=>$b}(keys %head_hash_old)) gt 0 ) || ( (grep lc($head_hash_old{$_}) ~~ /bezeichnung/ , sort{$a<=>$b}(keys %head_hash_old)) gt 0 ) )
							{
								$tmp_skiprow=1;
							}
						else
							{
								$skip_row++;
							}
					}
				print "#                                                    #\n";
				print "#    es werden die ersten ",$skip_row," Zeilen übersprungen\n";
				print "#                                                    #\n";
				#
				# for first worksheet
				#
				if ( $worksheet_i == 1 )
					{
						#
						#  substitute row names
						#
						print "#                                                    #\n";
						print "#   folgende Spaltenbeschriftungen wurden erkannt:   #\n";
						print "#                                                    #\n";
						print "#   | ";
						foreach( sort{$a<=>$b}(keys %head_hash_old) )
							{
								print $head_hash_old{$_}," | ";
							}
						print "\n";
						print "#                                                    #\n";
						print "#                                                    #\n";
						print "#   Bitte geben Sie die korrekten Spaltennamen       #\n";
						print "#   ein (-h:Hilfe):                                  #\n";
						my $tmp_var=0;
						#
						while( $tmp_var lt scalar(keys %head_hash_old) )
							{
								#
								print "#                   ",$head_hash_old{(sort{$a<=>$b}(keys %head_hash_old))[$tmp_var]},":";
								my $tmp_input = <STDIN>;
								chop($tmp_input);
								#
								if ( $tmp_input ~~ /-h/ )
									{
										# display help
										&list_tablearg(\%tableargs_hash,\%tabletrans_hash);
									}
								else
									{
										if ( defined($tmp_input) )
											{
												given( $tmp_input )
													{
														when( /\D/ )
															{ 
																# validate input
																if ( grep $tableargs_hash{$_} eq $tmp_input , sort{$a<=>$b}(keys %tableargs_hash) )
																	{
																		# store input
																		$head_hash_new{(sort{$a<=>$b}(keys %head_hash_old))[$tmp_var]} = $tmp_input;
																		$head_col_pos{$tmp_var} = $tmp_input;
																	}
																else
																	{
																		print "#                                                    #\n";
																		print "#   Eingabe wurde nicht in der Liste der Salten-     #\n";
																		print "#   beschriftungen gefunden! Bitte noch einmal       #\n";
																		print "#   versuchen (für eine Übersicht: -h)...            #\n";
																		print "#                                                    #\n";
																		$tmp_var--;
																	}
															}
														when( /^\d+\z/ )
															{ 
																# validate input
																if ( defined($tableargs_hash{$tmp_input}) )
																	{
																		# store value from list(hash)
																		$head_hash_new{(sort{$a<=>$b}(keys %head_hash_old))[$tmp_var]} = $tableargs_hash{$tmp_input};
																		$head_col_pos{$tmp_var} = $tableargs_hash{$tmp_input};
																	}
																else
																	{
																		print "#                                                    #\n";
																		print "#   Eingabe wurde nicht in der Liste der Salten-     #\n";
																		print "#   beschriftungen gefunden! Bitte noch einmal       #\n";
																		print "#   versuchen (für eine Übersicht: -h)...            #\n";
																		print "#                                                    #\n";
																		$tmp_var--;
																	}
															}
													}
												#
											}
										$tmp_var++;
									}
							}
						#
						print "#                                                    #\n";
						print "#   es werden folgende neue Spaltenbeschriftungen    #\n";
						print "#   verwendet:                                       #\n";
						print "#   | ";
						#
						foreach(sort{$a<=>$b}(keys %head_hash_new))
							{
								print $head_hash_new{$_}," | ";
							}
						print "\n";
						#
						# search for given prices
						# (if sellprice and listprice is given, then do not clone!)
						#
						if ( ( $input_hash{'clone_price'} ne 0 ) && ( ( (grep lc($head_hash_new{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %head_hash_new)) lt 1 ) || (grep lc($head_hash_new{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %head_hash_new)) lt 1 ) )
							{
								if ( (grep lc($head_hash_new{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %head_hash_new)) gt 0  )
									{
										print "#   Einfügen der Spalte \"listprice\"                #\n";
										$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'listprice';
										# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
										#            =>wrong row is choosen, if some rows are inserted automatically
									}
								elsif( (grep lc($head_hash_new{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %head_hash_new)) gt 0  )
									{
										print "#   Einfügen der Spalte \"sellprice\"                #\n";
										$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'sellprice';
										# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
										#            =>wrong row is choosen, if some rows are inserted automatically
									}
							}
						else
							{
								# if both prices given by default
								$input_hash{'clone_price'} = 0;
							}
						#
						print "#                                                    #\n";
						print "#                                                    #\n";	
						# 
						#
						# get value for "X"
						#
						# find out, if a value for "_X" is missing
						if ( ( (grep $head_hash_new{$_} ~~ /_X/ , sort{$a<=>$b}(keys %head_hash_new)) gt 0 ) ) 
							{
								print "#                                                    #\n";
								print "#   Bitte geben Sie den Lieferanten für folgende     #\n";
								print "#   Spalten an | ";
								foreach( (grep $head_hash_new{$_} ~~ /_X/ , sort{$a<=>$b}(keys %head_hash_new)) )
									{
										if ( !defined($head_hash_new{$_}) )
											{
												print $head_hash_new{$_}." | ";
											}
									}
								print "\n#   (Nummer/ Name/ -h):";
								my $tmp_input = <STDIN>;
								# remove "return/ new line"
								$tmp_input =~ s/\r|\n//g;
								#
								if ( defined($tmp_input) && grep(/-h/,$tmp_input) )
									{
										# print help
										&list_suppliers;
										# reset of former value, to get same input again
										$tmp_var--;
									}
								else
									{
										given( $tmp_input )
											{
												when( /\D/ )
													{ 
														# store input
														print "#                                                    #\n";
														print "#   Bitte die laufende Nummer für den Hersteller/    #\n";
														print "#   Lieferanten eingeben:                            #\n";
														&list_suppliers;
														# reset of former value, to get same input again
														$tmp_var--;
													}
												when( /^\d+\z/ )
													{ 
														# store value from list(hash)
														foreach( (grep $head_hash_new{$_} ~~ /_X/ , sort{$a<=>$b}(keys %head_hash_new)) )
															{
																$head_hash_new{$_} = (split(/_/,$head_hash_new{$_}))[0]."_".$tmp_input;
															}
														#
													}
											}
									}
								#
							}
						#
						# add "make_X", if needed
						#
						if ( defined($input_hash{'make_X'}) && ( (grep $head_hash_new{$_} ~~ /make_/ , sort{$a<=>$b}(keys %head_hash_new)) eq 0 ) )
							{
								print "#   Einfügen der Spalte \"make_".$input_hash{'make_X'}."\"\n";
								$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'make_'.$input_hash{'make_X'};
							}
						#
						# add "model_X", if needed
						#
						if ( defined($input_hash{'model_X'}) && ( (grep $head_hash_new{$_} ~~ /model_/ , sort{$a<=>$b}(keys %head_hash_new)) eq 0 ) )
							{
								print "#   Einfügen der Spalte \"model_".$input_hash{'model_X'}."\"\n";
								$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'model_'.$input_hash{'model_X'};
							}
						#
						# add "lastcost", if needed
						#
						if ( defined($input_hash{'lastcost'}) && ( (grep $head_hash_new{$_} eq 'lastcost' , sort{$a<=>$b}(keys %head_hash_new)) eq 0 ) )
							{
								print "#   Einfügen der Spalte \"lastcost\"\n";
								$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'lastcost';
							}
						#
						# add "lastcost_X", if needed
						#
						if ( defined($input_hash{'lastcost_X'}) && ( (grep $head_hash_new{$_} ~~ /lastcost_/ , sort{$a<=>$b}(keys %head_hash_new)) eq 0 ) )
							{
								print "#   Einfügen der Spalte \"lastcost_".$input_hash{'lastcost_X'}."\"\n";
								$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'lastcost_'.$input_hash{'lastcost_X'};
							}
						#
						# add "price_factor", if needed
						#
						if ( (grep $head_hash_new{$_} ~~ /price_factor/ , sort{$a<=>$b}(keys %head_hash_new)) eq 0 )
							{
								print "#   Einfügen der Spalte \"price_factor\"\n";
								$head_hash_new{ ((int((sort{$a<=>$b}(keys(%head_hash_new)))[-1]))+1000) } = 'price_factor';
							}
						#
						# translate column positions into "Kivitendo"- numbers
						#
						foreach(keys %head_hash_new)
							{
								if ( $_ < 1000)
									{
										my $tmpvar = $_;
										$head_hash_out{(grep $tableargs_hash{$_} eq $head_hash_new{$tmpvar} , keys %tableargs_hash)[0]} = $head_hash_new {$tmpvar};
									}
								else
									{
										$head_hash_out{$_} = $head_hash_new {$_};
									}
							}
						#
						# end of head creation
						#
						#######################################
						#
						# store head line in error_hash
						#
						foreach( sort{$a<=>$b}(keys(%head_hash_out)) )
							{
								if ( $head_hash_out{$_} !~ /rabattcode/ )
									{
										if ( $_ eq (sort{$a<=>$b}(keys(%head_hash_out)))[0] )
											{
												# first entry (without separator)
												$error_out_hash{'0'} = $head_hash_out{$_};
											}
										else
											{
												# normal "add"
												$error_out_hash{'0'} = $error_out_hash{'0'}."".$input_hash{'separator'}."".$head_hash_out{$_};
											}
									}
							}
					}
				#
				# create new file
				#
				$out_putadd = newfile(\%head_hash_out,\%input_hash,$out_putadd);
				#
				#
				# data calculation/ output
				#
				#
				# for all worksheets
				#
				my @column;
				my %column_hash;
				my $tmp_rownum=0;
				my $tmp_outvar=0;
				my $tmp_countvar=0;
				my $tmp_pn_pos=0;
				#
				for my $row ( $row_min+$skip_row+1 .. $row_max )
					{
						if( $tmp_rownum eq ($input_hash{'#maxlines'}-1) && $input_hash{'#maxlines'} !~ 0 )
							{
								$tmp_rownum=0;
								$input_hash{'output_filenum'}++;
								$out_putadd = newfile(\%head_hash_out,\%input_hash,$out_putadd);
							}
						else
							{
								$tmp_rownum++;
								
							}
						#
						# delete columm_hash and prevent "uninitialized value" errors
						foreach(keys %head_hash_out)
							{
								$column_hash{$_} = '';
							}
						for my $col ( $col_min .. $col_max ) 
							{
								if ( $input_hash{'input_format'} eq 'xls' )
									{
										#
										my $cell = $worksheet->get_cell( $row, $col );
										next unless $cell;
			               						# for each column in actual row: get unformatted cell value utf8 encoded
										#$column[$col] = uri_unescape( Encode::encode( "utf8",$cell->unformatted() ));
										# for each column in actual row: get unformatted cell value (no need to encode utf8, if binmode is set for output file!)
										#$column[$col] = uri_unescape( $cell->unformatted() );
										#
										if ( defined($head_hash_new{$col}) )
											{
												$column_hash{ (grep $tableargs_hash{$_} eq $head_hash_new{$col} , keys %tableargs_hash)[0] } = del_spaces( uri_unescape( $cell->unformatted() ) );
											}
									}
								elsif ( $input_hash{'input_format'} eq 'ods' )
									{
										#
										my $cell = $worksheet->get_cell( $row, $col )->get_text;
										next unless $cell;
										#$column[$col] = uri_unescape( Encode::encode( "utf8",$cell ));
										# no need to Encode/decode; seems that most ods are already utf8?
										$column_hash{ (grep $tableargs_hash{$_} eq $head_hash_new{$col} , keys %tableargs_hash)[0] } = del_spaces( uri_unescape( $cell ) );
									}
								elsif ( $input_hash{'input_format'} eq 'xlsx' )
									{
										#
										if ( defined($head_col_pos{$col}) )
											{
												my $cell = $worksheet-> {Cells} [$row] [$col];
												next unless $cell;
												$column_hash{ (grep $tableargs_hash{$_} eq $head_col_pos{$col} , keys %tableargs_hash)[0] } = del_spaces(  uri_unescape( $cell->value() ) );
											}
									}
								#
								#
								#
								if ( grep($head_hash_out{$col},'rabattcode') || defined($discount_hash{'global'}) )
									{
												#
												if ( !defined($head_hash_out{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]}) )
													{
														# insert "rabattcode" into haed_hash
														$head_hash_out{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} = 'rabattcode';
													}
												if ( scalar(keys %discount_hash) gt 0 && !defined($discount_hash{'global'}) )
													{
														if ( defined( $discount_hash{$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]}} ))
															{
																$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} = $discount_hash{$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]}};
															}
														else
															{
																$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} = 0;
															}
													}
												elsif  ( scalar(keys %discount_hash) gt 0  && defined($discount_hash{'global'}) )
													{
														$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} = $discount_hash{'global'};
													}
												else
													{
														$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} = 0;
													}
												#
												# substitute decimal sign
												# check, if "." is already there, like: 1.000,00 => do not change "." to ","!
												if ( grep(/\,/,$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]}) )
													{
														# delete points (only decimal point should be left) 
														# => needed for calculation of prices
														$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} =~ s/\.//g;
														# komma to point
														$column_hash{(grep $tableargs_hash{$_} ~~ /rabattcode/ , keys %tableargs_hash)[0]} =~ s/\,/\./g;
													}
											}
								#
							}
						#
						# call "parse_n_write"
						if ( defined(my $error = parse_n_write(\%column_hash,\%head_hash_out,\%input_hash,$out_putadd)) )
							{
								# fill error hash (if error occurs)
								if ( length($error) gt 0 )
									{
										$error_out_hash{((int((sort{$a<=>$b}(keys(%error_out_hash)))[-1]))+1)} = $error;
									}
							}
						#
						# output to console
						#
						print "#   Fortschritt => verarbeite Eintrag: ".$row."\r";
						#last if $. = 1;
					}
				print "\n";
			}
		#
		# error output into error file
		#
		error_out(\%error_out_hash,\$input_hash{'output_file'});
	}
#
#############################################################################################
#                                                                                           #
#                parse ascii file (Kramp, CNH, DeLaval, ....)                               #
#                                                                                           #
#############################################################################################
#
sub ascii
	{
		my %input_args_hash = %{(shift)};
		my %tableargs_hash = %{(shift)};
		my %tabletrans_hash = %{(shift)};
		my %suppliers_list = %{(shift)};
		my %suppabbr_list = %{(shift)};
		my %unit_trans = %{(shift)};
		my %output_hash;
		my %error_output_hash;				# hash containing errors (errornous output lines, to be written in separate file)
		my %head_hash;
		my %head_def_hash;
		my %head_trans_hash;
		my %discount_hash;
		my $out_putadd;
 		my %errorout_hash;
		#
		if (`file $input_args_hash{'input_file'}` !~ /text/)
			{
				print "#                                                    #\n";
				print "#    Fehler: Typ der Preislistendatei nicht erkannt! #\n";
				print "#            Es muss eine Datei vom Typ \"Text\"       #\n";
				print "#            angegeben werden!                       #\n";
				print "#                                ==> ABBRUCH!        #\n";
				print "#                                                    #\n";
				print "######################################################\n";
				exit 2;
				
			}
		else
			{
				#
				# read column definitions (names, position and lenght) of text file
				#
				my $def_file;
				open($def_file,'<:encoding(UTF-8)','config/head_'.$input_args_hash{'make_X'}.'_def.txt') or die "Could not open $def_file: $!";
				while( my $line = <$def_file> )
					{
						# read col definitions of given file
						if (!grep(/#/,substr($line,0,10)) )
							{
								$head_def_hash{(split(/;;/,$line))[0]} = [(split(/;;/,$line))[1],(split(/;;/,$line))[2]];
							}
					}
				close $def_file;
				open($def_file,'<:encoding(UTF-8)','config/head_'.$input_args_hash{'make_X'}.'_trans.txt') or die "Could not open $def_file: $!";
				while( my $line = <$def_file> )
					{
						# read col definitions of given file
						if (!grep(/#/,substr($line,0,10)) )
							{
								$head_trans_hash{del_spaces((split(/;;/,$line))[0])} = del_spaces((split(/;;/,$line))[1]);
							}
					}
				close $def_file;
				if ( -r 'config/'.$input_args_hash{'make_X'}.'_discount.txt' )
					{
						print "#                                                    #\n";
						print "#   verwende Rabattcodedatei: \"config/".$input_args_hash{'make_X'}."_discount.txt\"\n";
						print "#                                                    #\n";
						open($def_file,'<:encoding(UTF-8)','config/'.$input_args_hash{'make_X'}.'_discount.txt') or die "Could not open $def_file: $!";
						while( my $line = <$def_file> )
							{
								# read discount value
								$discount_hash{(split(/;;/,$line))[0]} = del_spaces((split(/;;/,$line))[1]);
							}
						close $def_file;
					}
				elsif ( -r $input_args_hash{'glob_discount'} )
					{
						print "#                                                    #\n";
						print "#   verwende Rabattcodedatei: \"".$input_args_hash{'glob_discount'}."\"\n";
						print "#                                                    #\n";
						open($def_file,'<:encoding(UTF-8)',$input_args_hash{'glob_discount'}) or die "Could not open $def_file: $!";
						while( my $line = <$def_file> )
							{
								# read discount value
								if( (split(/;;/,$line))[0] eq $input_args_hash{'make_X'} )
									{
										$discount_hash{'global'} = del_spaces((split(/;;/,$line))[1]);
										print "#                                                    #\n";
										print "#   anzuwendender Rabatt (global): ".$discount_hash{'global'}."\%\n";
										print "#                                                    #\n";
									}
							}
						close $def_file;
					}
				print "#                                                    #\n";
				print "#   Erstellen der Tabellenspaltenbeschriftung...     #\n";
				print "#                                                    #\n";
				foreach(keys %head_def_hash)
					{
						if ( defined($head_trans_hash{$_}) )
							{
								if ( length($head_trans_hash{$_}) gt 0  )
									{
										# translate column names in Kivitendo ones
										$head_hash{$head_trans_hash{$_}} = $tableargs_hash{$head_trans_hash{$_}};
									}
							}
					}
				print "#                                                    #\n";
				print "#   Es werden folgende Spalten ausgelesen:           #\n";
				print "#                                                    #\n";
				print "#   | ";
				foreach(sort{$a<=>$b}(keys %head_hash))
					{
						print $head_hash{$_}." | ";
					}
				print "\n";
				print "#                                                    #\n";
				#
				# search for given prices
				# (if sellprice and listprice is given, then do not clone!)
				#
				if ( ( $input_args_hash{'clone_price'} ne 0 ) && ( ( (grep lc($head_hash{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %head_hash)) lt 1 ) || (grep lc($head_hash{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %head_hash)) lt 1 ) )
					{
						if ( (grep lc($head_hash{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %head_hash)) gt 0  )
							{
								print "#   Einfügen der Spalte \"listprice\"                  #\n";
								$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'listprice';
								# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
								#            =>wrong row is choosen, if some rows are inserted automatically
							}
						elsif( (grep lc($head_hash{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %head_hash)) gt 0  )
							{
								print "#   Einfügen der Spalte \"sellprice\"                  #\n";
								$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'sellprice';
								# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
								#            =>wrong row is choosen, if some rows are inserted automatically
							}
					}
				else
					{
						# if both prices given by default
						$input_args_hash{'clone_price'} = 0;
					}
				#
				# add "make_X", if needed
				#
				if ( defined($input_args_hash{'make_X'}) && ( (grep $head_hash{$_} ~~ /make_/ , sort{$a<=>$b}(keys %head_hash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"make_".$input_args_hash{'make_X'}."\"                     #\n";
						$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'make_'.$input_args_hash{'make_X'};
					}
				#
				# add "model_X", if needed
				#
				if ( defined($input_args_hash{'model_X'}) && ( (grep $head_hash{$_} ~~ /model_/ , sort{$a<=>$b}(keys %head_hash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"model_".$input_args_hash{'model_X'}."\"                    #\n";
						$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'model_'.$input_args_hash{'model_X'};
					}
				#
				# add "lastcost", if needed
				#
				if ( defined($input_args_hash{'lastcost'}) && ( (grep $head_hash{$_} eq 'lastcost' , sort{$a<=>$b}(keys %head_hash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"lastcost\"                   #\n";
						$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'lastcost';
					}
				#
				# add "lastcost_X", if needed
				#
				if ( defined($input_args_hash{'lastcost_X'}) && ( (grep $head_hash{$_} ~~ /lastcost_/ , sort{$a<=>$b}(keys %head_hash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"lastcost_".$input_args_hash{'lastcost_X'}."\"                 #\n";
						$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'lastcost_'.$input_args_hash{'lastcost_X'};
					}
				#
				# add "price_factor", if needed
				#
				if ( (grep $head_hash{$_} ~~ /price_factor/ , sort{$a<=>$b}(keys %head_hash)) eq 0 )
					{
						print "#   Einfügen der Spalte \"price_factor\"               #\n";
						$head_hash{ ((int((sort{$a<=>$b}(keys(%head_hash)))[-1]))+100) } = 'price_factor';
					}
				#
				# find position of last value and calculate min. required string length
				#
				$input_args_hash{'min_length'} = 0;
				$input_args_hash{'min_length'} +=0;
				foreach( keys %head_def_hash )
					{
						if ( $input_args_hash{'min_length'} < ($head_def_hash{$_}[0]+$head_def_hash{$_}[1]-1) )
							{
								# set min. requiered line length
								$input_args_hash{'min_length'} = ($head_def_hash{$_}[0]+$head_def_hash{$_}[1]-1);
							}
					}
				#
				# end of head creation
				#
				# create new output file
				$out_putadd = newfile(\%head_hash,\%input_args_hash,$out_putadd);
				$input_args_hash{'output_filenum'}++;
				#
				# prepare output hash to supress "use of unintialized value" warnings
				#
				foreach(sort{$a<=>$b}(keys %head_hash))
					{
						$output_hash{$_} = '';
					}
				#######################################
				#
				# store head line in error_hash
				#
				foreach( sort{$a<=>$b}(keys(%head_hash)) )
					{
						if ( $head_hash{$_} ne 'rabattcode' )
							{
								if ( $_ eq (sort{$a<=>$b}(keys(%head_hash)))[0] )
									{
										# first entry (without separator)
										$errorout_hash{'0'} = $head_hash{$_};
									}
								else
									{
										# normal "add"
										$errorout_hash{'0'} .= $input_args_hash{'separator'};
										$errorout_hash{'0'} .= $head_hash{$_};
									}
							}
					}
				print "#                                                    #\n";
				print "#   Beginn des Auslesens der Textdatei.......        #\n";
				print "#                                                    #\n";
				#
				my $charset;
				if ( grep(/iso/,(split(/=/,`file -i -b $input_args_hash{'input_file'}`))[1]) || grep(/utf/,(split(/=/,`file -i -b $input_args_hash{'input_file'}`))[1]) )
					{
						$charset = (split(/=/,`file -i -b $input_args_hash{'input_file'}`))[1];
					}
				elsif ( grep(/unknown-8bit/,(split(/=/,`file -i -b $input_args_hash{'input_file'}`))[1]) )
					{
						$charset = 'IBM850';
					}
				else
					{
						print "#                                                    #\n";
						print "#    Fehler: Charset/ Codierung der Preislistendatei #\n";
						print "#            wurde nicht erkannt!                    #\n";
						print "#            Die Datei weist folgende Kodierung auf:  \n";
						print "#              ".`file -i -b $input_args_hash{'input_file'}`."\n";
						print "#                                                    #\n";
						print "#                                ==> ABBRUCH!        #\n";
						print "#                                                    #\n";
						print "######################################################\n";
						exit 3;
					}
				open(my $inputfile,'<:encoding('.$charset.')',$input_args_hash{'input_file'});
				while( my $line = <$inputfile>)  
					{
						# split files, if max line numbers reached
						if( $. eq $input_args_hash{'output_filenum'}*$input_args_hash{'#maxlines'} )
							{
								$out_putadd = newfile(\%head_hash,\%input_args_hash,$out_putadd);
								$input_args_hash{'output_filenum'}++;
							}
						#
						# test, if current line is long enouth
						if ( length($line) >= $input_args_hash{'min_length'} )
							{
								foreach(keys %head_hash)
									{
										my $tmp_key = $_;
										#print "tmp_key:".$tmp_key."=>".(grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]."\n";
										if ( defined($head_hash{$tmp_key}) )
											{
												#
												if ( length($head_hash{$tmp_key}) gt 0  && $head_hash{$tmp_key} ne 'rabattcode' && $head_hash{$tmp_key} ne 'listprice' && $head_hash{$tmp_key} ne 'listprice' && $head_hash{$tmp_key} !~ /lastcost/ && $head_hash{$tmp_key} !~ /weight/ )
													{
														# set output hash, if column position is defined
														if ( defined((grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]) )
															{
																$output_hash{$tmp_key} =  del_spaces(substr($line,$head_def_hash{(grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]}[0]-1,$head_def_hash{(grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]}[1]));
															}
														#
													}
												elsif ( length($head_hash{$tmp_key}) gt 0  && $head_hash{$tmp_key} ne 'rabattcode' && ( $head_hash{$tmp_key} eq 'listprice' || $head_hash{$tmp_key} eq 'listprice' || $head_hash{$tmp_key} ~~ /lastcost/ || $head_hash{$tmp_key} ~~ /weight/ ) )
													{
														# set output hash for prices and insert decimal point
														if ( defined((grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]) )
															{
																$output_hash{$tmp_key} =  del_spaces(substr($line,$head_def_hash{(grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]}[0]-1,$head_def_hash{(grep $head_trans_hash{$_} eq $tmp_key , keys %head_trans_hash)[0]}[1]));
																# insert decimal sign
																substr($output_hash{$tmp_key},-2,0) = ',';
																# delete leading zeros
																$output_hash{$tmp_key} =~ s/^(-?)0*/$1/;
																# check, if first sign is number
																if ( substr($output_hash{$tmp_key},0,1) eq ',' )
																	{
																		# insert one (!) leading zereo, if first character is decimal sign
																		$output_hash{$tmp_key} = "0".$output_hash{$tmp_key};
																	}
															}
														#
													}
												elsif ( length($head_hash{$tmp_key}) gt 0  && $head_hash{$tmp_key} eq 'rabattcode')
													{
														#
														if ( scalar(keys %discount_hash) gt 0 && !defined($discount_hash{'global'}) )
															{
																if ( defined( $discount_hash{del_spaces(substr($line,$head_def_hash{(grep $head_trans_hash{$_} eq '1000' , keys %head_trans_hash)[0]}[0]-1,$head_def_hash{(grep $head_trans_hash{$_} eq '1000' , keys %head_trans_hash)[0]}[1]))} ))
																	{
																		$output_hash{'1000'} = $discount_hash{del_spaces(substr($line,$head_def_hash{(grep $head_trans_hash{$_} eq '1000' , keys %head_trans_hash)[0]}[0]-1,$head_def_hash{(grep $head_trans_hash{$_} eq '1000' , keys %head_trans_hash)[0]}[1]))};
																	}
																else
																	{
																		$output_hash{'1000'} = 0;
																	}
															}
														elsif  ( scalar(keys %discount_hash) gt 0  && defined($discount_hash{'global'}) )
															{
																$output_hash{'1000'} = $discount_hash{'global'};
															}
														else
															{
																$output_hash{'1000'} = 0;
															}
														#
														# substitute decimal sign
														# check, if "." is already there, like: 1.000,00 => do not change "." to ","!
														if ( grep(/\,/,$output_hash{'1000'}) )
															{
																# delete points (only decimal point should be left) 
																# => needed for calculation of prices
																$output_hash{'1000'} =~ s/\.//g;
																# komma to point
																$output_hash{'1000'} =~ s/\,/\./g;
															}
													}
											}
									}
								#
								# delete leading zeros in all values, except Text!
								#
								foreach(keys %output_hash)
									{
										if ( $head_hash{$_} ~~ /price/ || $head_hash{$_} ~~ /unit/ || $head_hash{$_} ~~ /lastcost/  )
											{
												$output_hash{$_}  =~ s/^(-?)0*/$1/;
											}
									}
								# call "parse_n_write"
								if ( defined(my $error = parse_n_write(\%output_hash,\%head_hash,\%input_args_hash,$out_putadd)) )
									{
										# fill error hash (if error occurs)
										if ( length($error) gt 0 )
											{
												$errorout_hash{((int((sort{$a<=>$b}(keys(%errorout_hash)))[-1]))+1)} = $error;
											}
									}
							}
						else
							{
								$errorout_hash{((int((sort{$a<=>$b}(keys(%errorout_hash)))[-1]))+1)} = "ERROR_LENGTH:".$line."\n";
							}
						#
						print "#   Fortschritt => verarbeite Datensatz: ".$.."\r";
						# stop parsing (for debug purpose)
						#last if $. == 10;
					}
				#
				print "#   Es wurden ".$.." Datensätze verarbeitet                                 \n";
				#
				close $inputfile;
				#
				# error output into error file
				#
				error_out(\%errorout_hash,\$input_args_hash{'output_file'});
			}
	}
#
#############################################################################################
#                                                                                           #
#                parse data in DATANORM 5.0 format                                          #
#                                                                                           #
#############################################################################################
#
sub datanorm_50
	{
		my %inputargshash = %{(shift)};
		my %tableargshash = %{(shift)};
		my %tabletranshash = %{(shift)};
		my %supplierslist = %{(shift)};
		my %suppabbrlist = %{(shift)};
		my %unittrans = %{(shift)};
		my %char_set =%{(shift)};
		my %outputhash;
		my $outputadd;
		my %erroroutputhash;				# hash containing errors (errornous output lines, to be written in separate file)
		my %headhash;
		my %headdefhash;
		my %headtranshash;
		my %discounthash;
		my %discountdefhash;
		my %errorouthash;
 		my %datanorm_files_hash;
 		#
 		# default_values
 		#
 		$inputargshash{'dn50_discount_separator'} = ';';
 		$inputargshash{'dn50_separator'} = ';';
		$inputargshash{'dn50_discount_div'} = 1;
		$inputargshash{'dn50_price_div'} = 1;
		$inputargshash{'dn50_parts_keyw'} = 'Artikel';
		$inputargshash{'dn50_group_keyw'} = 'Warengruppen';
		$inputargshash{'dn50_discount_keyw'} = 'Rabatt';
		$inputargshash{'dn50_parts_key'} = 'A';
		$inputargshash{'dn50_group_key'} = 'S';
		$inputargshash{'dn50_discount_key'} = 'R';
		$inputargshash{'dn50_version_key'} = 'V';
		#
		$tabletranshash{'dn50_obsolete_new'} = '0';
		$tabletranshash{'dn50_obsolete_renew'} = '0';
		$tabletranshash{'dn50_obsolete_delete'} = '1';
 		#
		# read column definitions (names, position and lenght) of text file
		#
		my $def_file;
		open($def_file,'<:encoding(UTF-8)','config/head_'.$inputargshash{'make_X'}.'_def.txt') or die "Could not open $def_file: $!";
		while( my $line = <$def_file> )
			{
				# read col definitions of given file
				if (!grep(/#/,substr($line,0,10)) )
					{
						if ( (split(/;;/,$line))[0] ~~ /part/ )
							{
								# if line contains part column defintion
								$headdefhash{(split(/;;/,$line))[1]} = del_spaces((split(/;;/,$line))[2]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /separator/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_separator'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /discount_separator/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_discount_separator'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /price_div/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_price_div'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /discount_div/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_discount_div'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /dn50_parts_keyw/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_parts_keyw'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /dn50_group_keyw/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_group_keyw'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /dn50_discount_keyw/ )
							{
								# if line contains separator information for datanorm file
								$inputargshash{'dn50_discount_keyw'} = del_spaces((split(/;;/,$line))[1]);
							}
						elsif( (split(/;;/,$line))[0] ~~ /dn50_obsolete/ )
							{
								# if line contains separator information for datanorm file
								if ( (split(/;;/,$line))[1] ~~ /new/ && (split(/;;/,$line))[1] !~ /renew/ )
									{
										$inputargshash{'dn50_obsolete_new'} = del_spaces((split(/;;/,$line))[2]);
									}
								elsif ( (split(/;;/,$line))[1] ~~ /renew/ )
									{
										$inputargshash{'dn50_obsolete_renew'} = del_spaces((split(/;;/,$line))[2]);
									}
								elsif ( (split(/;;/,$line))[1] ~~ /delete/ )
									{
										$inputargshash{'dn50_obsolete_delete'} = del_spaces((split(/;;/,$line))[2]);
									}
							}
					}
			}
		close $def_file;
		#
		open($def_file,'<:encoding(UTF-8)','config/head_'.$inputargshash{'make_X'}.'_trans.txt') or die "Could not open $def_file: $!";
		while( my $line = <$def_file> )
			{
				# read col definitions of given file
				if (!grep(/#/,substr($line,0,10)) )
					{
						if ( (split(/;;/,$line))[0] ~~ /part/ )
							{
								$headtranshash{del_spaces((split(/;;/,$line))[1])} = del_spaces((split(/;;/,$line))[2]);
							}
					}
			}
		close $def_file;
		#
		# check, if "input_file" is an directory
		#
		if ( !-d $inputargshash{'input_file'} )
			{
				#
				# if an file is given...
				#
				$inputargshash{'input_file'} =~ s/\/\//\//g;
				my @tmp_dir = split(/\//,$inputargshash{'input_file'});
				$inputargshash{'input_file'} = '';
				#
				# delete last element (file name)
				pop(@tmp_dir);
				foreach(@tmp_dir)
					{
						$inputargshash{'input_file'} .= $_."/";
					}
				print "#                                                    #\n";
				print "#      verwende Verzeichnis:".$inputargshash{'input_file'}."\n";
				print "#                                                    #\n";
			}
		#
		#
		#
		my $tmp_dir;
		opendir($tmp_dir,$inputargshash{'input_file'});
		#
		# scan of input directory
		#
		foreach(sort(readdir($tmp_dir)))
			{
				my $tmp_direntry = $_;
				#
				# search for files
				#
				if ( ( $tmp_direntry !~ /\.\./ || ( $tmp_direntry !~ /\./ ) && length($tmp_direntry) eq 1 ) && `file $inputargshash{'input_file'}$tmp_direntry` ~~ /text/ )
					{
						#
						# read first line for file indentification
						#
						open(my $tmp_file,'<:encoding('.$char_set{$inputargshash{'make_X'}}.')',$inputargshash{'input_file'}."".$tmp_direntry);
						#
						# read first line for identification of input file
						#
						chomp(my $l_ine = <$tmp_file>);
						#
						if ( (split(/\;/,$l_ine))[0] ~~ /$inputargshash{'dn50_version_key'}/ )
							{
								#
								# identification of files by keyword in first line of current file
								#
								if ( grep(/$inputargshash{'dn50_parts_keyw'}/,$l_ine)  )
									{
										$inputargshash{'dn50_partsfile'} = $tmp_direntry;
									}
								elsif ( grep(/$inputargshash{'dn50_discount_keyw'}/,$l_ine)  )
									{
										$inputargshash{'dn50_discountfile'} = $tmp_direntry;
									}
								elsif ( grep(/$inputargshash{'dn50_group_keyw'}/,$l_ine)  )
									{
										$inputargshash{'dn50_partsgroupfile'} = $tmp_direntry;
									}
							}
						#
						close($tmp_file);
					}
			}
		closedir($tmp_dir);
		#
		print "#                                                    #\n";
		if ( defined($inputargshash{'dn50_partsfile'}) )
			{
				print "#      Datanorm- Artikeldatei:".$inputargshash{'dn50_partsfile'}."\n";
			}
		if ( defined($inputargshash{'dn50_partsgroupfile'}) )
			{
				print "# Datanorm- Warengruppendatei:".$inputargshash{'dn50_partsgroupfile'}."\n";
			}
		if ( defined($inputargshash{'dn50_discountfile'}) )
			{
				print "#       Datanorm- Rabattdatei:".$inputargshash{'dn50_discountfile'}."\n";
			}
		#
		if ( defined($inputargshash{'dn50_discountfile'}) )
			{
				print "#                                                    #\n";
				print "#       Einlesen der Rabattcodes.....                #\n";
				#
				open(my $tmp_file,'<:encoding('.$char_set{$inputargshash{'make_X'}}.')',$inputargshash{'input_file'}."".$inputargshash{'dn50_discountfile'});
				while(my $l_ine = <$tmp_file>)
					{
						if ( (split(/\$inputargshash{'dn50_discount_separator'}/,$l_ine))[0] ~~ /$inputargshash{'dn50_discount_key'}/)
							{
								$discounthash{(split(/$inputargshash{'dn50_discount_separator'}/,$l_ine))[1]} = (split(/$inputargshash{'dn50_discount_separator'}/,$l_ine))[3]/$inputargshash{'dn50_discount_div'};
							}
					}
				close($tmp_file);
			}
		#
		#print "#                                                    #\n";
		#print "#       Einlesen der Warengruppen.....               #\n";
		#if ( defined($inputargshash{'dn50_partsgroupfile'}) )
		#	{
		#		open(my $tmp_file,'<:encoding('.$char_set{$inputargshash{'make_X'}}.')',$inputargshash{'input_file'}."".$inputargshash{'dn50_partsgroupfile'});
		#		while(my $l_ine = <$tmp_file>)
		#			{
		#				if ( lc((split(/\;/,$l_ine))[0]) ~~ /r/)
		#					  {
		#						$partsgrouphash{(split(/\;/,$l_ine))[1]} = (split(/\;/,$l_ine))[3]/100;
		#					  }
		#			}
		#		close($tmp_file);
		#	}
		print "#                                                    #\n";
		print "#       Beginn der Verarbeitung der Artikel.....     #\n";
		print "#                                                    #\n";
		if ( !defined($inputargshash{'dn50_partsfile'}) )
			{
				print "#                                                    #\n";
				print "#    Fehler: DATANORM- Preislistendatei nicht        #\n";
				print "#            gefunden! Es muss ein Verzeichnis       #\n";
				print "#            mit min. einer Text- Datei, die Preise  #\n";
				print "#            enthält  mit angegeben werden!          #\n";
				print "#                                ==> ABBRUCH!        #\n";
				print "#                                                    #\n";
				print "######################################################\n";
				exit 2;
			}
		else
			{
				#
				# assemble head information
				#
				foreach( keys %headdefhash )
					{
						if ( defined($headtranshash{$_}) )
							{
								if ( $headtranshash{$_} gt 0 )
									{
										$headhash{$headtranshash{$_}} = $tableargshash{$headtranshash{$_}};
									}
							}
						#
					}
				#
				# search for given prices
				# (if sellprice and listprice is given, then do not clone!)
				#
				if ( ( $inputargshash{'clone_price'} ne 0 ) && ( ( (grep lc($headhash{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %headhash)) lt 1 ) || (grep lc($headhash{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %headhash)) lt 1 ) )
					{
						if ( (grep lc($headhash{$_}) ~~ /sellprice/ , sort{$a<=>$b}(keys %headhash)) gt 0  )
							{
								print "#   Einfügen der Spalte \"listprice\"                  #\n";
								$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'listprice';
								# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
								#            =>wrong row is choosen, if some rows are inserted automatically
							}
						elsif( (grep lc($headhash{$_}) ~~ /listprice/ , sort{$a<=>$b}(keys %headhash)) gt 0  )
							{
								print "#   Einfügen der Spalte \"sellprice\"                  #\n";
								$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'sellprice';
								# ATTENTION: INDEX- setting => last+100 due to problem, some columns are left
								#            =>wrong row is choosen, if some rows are inserted automatically
							}
					}
				else
					{
						# if both prices given by default
						$inputargshash{'clone_price'} = 0;
					}
				#
				# add "make_X", if needed
				#
				if ( defined($inputargshash{'make_X'}) && ( (grep $headhash{$_} ~~ /make_/ , sort{$a<=>$b}(keys %headhash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"make_".$inputargshash{'make_X'}."\"\n";
						$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'make_'.$inputargshash{'make_X'};
					}
				#
				# add "model_X", if needed
				#
				if ( defined($inputargshash{'model_X'}) && ( (grep $headhash{$_} ~~ /model_/ , sort{$a<=>$b}(keys %headhash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"model_".$inputargshash{'model_X'}."\"\n";
						$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'model_'.$inputargshash{'model_X'};
					}
				#
				# add "lastcost", if needed
				#
				if ( defined($inputargshash{'lastcost'}) && ( (grep $headhash{$_} eq 'lastcost' , sort{$a<=>$b}(keys %headhash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"lastcost\"\n";
						$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'lastcost';
					}
				#
				# add "lastcost_X", if needed
				#
				if ( defined($inputargshash{'lastcost_X'}) && ( (grep $headhash{$_} ~~ /lastcost_/ , sort{$a<=>$b}(keys %headhash)) eq 0 ) )
					{
						print "#   Einfügen der Spalte \"lastcost_".$inputargshash{'lastcost_X'}."\"\n";
						$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'lastcost_'.$inputargshash{'lastcost_X'};
					}
				#
				# add "price_factor", if needed
				#
				if ( (grep $headhash{$_} ~~ /price_factor/ , sort{$a<=>$b}(keys %headhash)) eq 0 )
					{
						print "#   Einfügen der Spalte \"price_factor\"\n";
						$headhash{ ((int((sort{$a<=>$b}(keys(%headhash)))[-1]))+100) } = 'price_factor';
					}
				#
				# prepare output hash to supress "use of unintialized value" warnings
				#
				foreach(sort{$a<=>$b}(keys %headhash))
					{
						$outputhash{$_} = '';
					}
				#######################################
				#
				# store head line in error_hash
				#
				foreach( sort{$a<=>$b}(keys(%headhash)) )
					{
						if ( $headhash{$_} ne 'rabattcode' )
							{
								if ( $_ eq (sort{$a<=>$b}(keys(%headhash)))[0] )
									{
										# first entry (without separator)
										$erroroutputhash{'0'} = $headhash{$_};
									}
								else
									{
										# normal "add"
										$erroroutputhash{'0'} .= $inputargshash{'separator'};
										$erroroutputhash{'0'} .= $headhash{$_};
									}
							}
					}
				print "#                                                    #\n";
				print "#   Beginn des Auslesens der Textdatei.......        #\n";
				print "#                                                    #\n";
				#
				# create new file
				#
				$outputadd = newfile(\%headhash,\%inputargshash,$outputadd);
				$inputargshash{'output_filenum'}++;
				#
				open(my $tmp_file,'<:encoding('.$char_set{$inputargshash{'make_X'}}.')',$inputargshash{'input_file'}."".$inputargshash{'dn50_partsfile'});
				while(my $l_ine = <$tmp_file>)
					{
						if ( (split(/\;/,$l_ine))[0] ~~ /A/ )
							{
								#
								print "#   Fortschritt => verarbeite Datensatz: ".$.."\r";
								foreach(split(/\;/,$l_ine))
									{
										foreach( keys %headtranshash )
											{
												if ( defined($headtranshash{$_}) )
													{
														if ( length($headtranshash{$_}) gt 0 )
															{
																$outputhash{$headtranshash{$_}} = (split(/\;/,$l_ine))[$headdefhash{$_}-1];
															}
													}	
											}
									}
								#
								# subsitute discount group by discount value
								#
								if ( (grep $headhash{$_} eq 'rabattcode' , keys %headhash)[0] gt 0 )
									{
										if ( defined($discounthash{$outputhash{(grep $headhash{$_} eq 'rabattcode' , keys %headhash)[0]}}) )
											{
												$outputhash{(grep $headhash{$_} eq 'rabattcode' , keys %headhash)[0]} = $discounthash{$outputhash{(grep $headhash{$_} eq 'rabattcode' , keys %headhash)[0]}};
											}
										else
											{
												$outputhash{(grep $headhash{$_} eq 'rabattcode' , keys %headhash)[0]} = 0;
											}
									}
								#
								# subsitute obsolete value by Kivitendo's one's
								#
								if ( (grep $headhash{$_} eq 'obsolete' , keys %headhash)[0] gt 0 )
									{
										if ( defined($outputhash{(grep $headhash{$_} eq 'obsolete' , keys %headhash)[0]}) )
											{
												foreach( (grep $inputargshash{$_} ~~ /$outputhash{(grep $headhash{$_} eq 'obsolete' , keys %headhash)[0]}/ , sort(keys %inputargshash)) )
													{
														if ( $_ ~~ /obsolete/ )
															{
																$outputhash{(grep $headhash{$_} eq 'obsolete' , keys %headhash)[0]} = $tabletranshash{$_};
															}
													}
											}
										else
											{
												$outputhash{(grep $headhash{$_} eq 'obsolete' , keys %headhash)[0]} = 'ERROR_OBSOLETE:'.$outputhash{(grep $headhash{$_} eq 'obsolete' , keys %headhash)[0]};
											}
									}
								#
								# insert missing keys in output-hash (to prevent "Use of uninitialized value within..." messages)
								#
								if ( (grep $headhash{$_} eq 'listprice' , keys %headhash)[0] gt 0 )
									{
										if ( !defined($outputhash{(grep $headhash{$_} eq 'listprice' , keys %headhash)[0]}) )
											{
												$outputhash{(grep $headhash{$_} eq 'listprice' , keys %headhash)[0]} = '';
											}
									}
								if ( (grep $headhash{$_} eq 'sellprice' , keys %headhash)[0] gt 0 )
									{
										if ( !defined($outputhash{(grep $headhash{$_} eq 'sellprice' , keys %headhash)[0]}) )
											{
												$outputhash{(grep $headhash{$_} eq 'sellprice' , keys %headhash)[0]} = '';
											}
									}
								if ( (grep $headhash{$_} eq 'lastcost' , keys %headhash)[0] gt 0 )
									{
										if ( !defined($outputhash{(grep $headhash{$_} eq 'lastcost' , keys %headhash)[0]}) )
											{
												$outputhash{(grep $headhash{$_} eq 'lastcost' , keys %headhash)[0]} = '';
											}
									}
								if ( (grep $headhash{$_} eq 'lastcost_'.$inputargshash{'make_X'} , keys %headhash)[0] gt 0 )
									{
										if ( !defined($outputhash{(grep $headhash{$_} eq 'lastcost_'.$inputargshash{'make_X'} , keys %headhash)[0]}) )
											{
												$outputhash{(grep $headhash{$_} eq 'lastcost_'.$inputargshash{'make_X'} , keys %headhash)[0]} = '';
											}
									}
								#
								# insert "comma" in price values, if source file contains "comma-less" prices
								#
								foreach( (grep $headhash{$_} ~~ /price/ , sort{$a<=>$b}(keys %headhash)) )
									{
										if ( defined($outputhash{$_}) && $headhash{$_} !~ /price_factor/ )
											{
												if ( length($outputhash{$_}) ne 0 )
													{
														$outputhash{$_} = $outputhash{$_}/$inputargshash{'dn50_price_div'};
													}
											}
									}
								# call "parse_n_write"
								if ( defined(my $error = parse_n_write(\%outputhash,\%headhash,\%inputargshash,$outputadd)) )
									{
										# fill error hash (if error occurs)
										if ( length($error) gt 0 )
											{
												$erroroutputhash{((int((sort{$a<=>$b}(keys(%erroroutputhash)))[-1]))+1)} = $error;
											}
									}
								elsif ( $. eq (($inputargshash{'#maxlines'}-1)*$inputargshash{'output_filenum'}) )
									{
										#
										# create new file
										#
										$outputadd = newfile(\%headhash,\%inputargshash,$outputadd);
										$inputargshash{'output_filenum'}++;
										#
									}
							}
						
						
					}
				print "#                                                    #\n";
				print "#   Es wurden ".$.." Datensätze verarbeitet                                 \n";
				#
				close($tmp_file);
				#
				# error output into error file
				#
				error_out(\%erroroutputhash,\$inputargshash{'output_file'});
			}		
	}
#
#############################################################################################
#                                                                                           #
#                parse data in PDF document format                                          #
#                                                                                           #
#############################################################################################
#
sub pdf
	{
		print "#                                                    #\n";
 		print "#   Die Verarbeitung von PDF- Dateien ist derzeit    #\n";
 		print "#   noch nicht möglich!                              #\n";
 		print "#   Sollte jemand helfen wollen:                     #\n";
 		print "#                 Hilfe ist gerne willkommen!        #\n";
 		print "#                                                    #\n";
 		#
 	}
#
#############################################################################################
#                                                                                           #
#                parse data and write to file                                               #
#                                                                                           #
#############################################################################################
#
sub parse_n_write
	{
		#
		###########################################################################################
		#                                                                                         #
		#  definition of output_hash:                                                             #
		#              %output_hash = { key_1 => value_1,                                         #
		#                               key_2 => value_2,                                         #
		#                               .......};                                                 #
		#                key_# : key is number of rowname in Kivitendo (ex.: '13' for listprice)  #
		#              value_# : value for key                                                    #
		#                                                                                         #
		# definition of head_hash_new:                                                            #
		#            %head_hash_new = { key_1 => value_1,                                         #
		#                               key_2 => value_2,                                         #
		#                               .......};                                                 #
		#                key_# : key is number of rowname in Kivitendo (ex.: '13' for listprice)  #
		#              value_# : rowname in Kivtendo (ex. 'listprice'; used for head output in    #
		#                        output files                                                     #
		#                                                                                         #
		#      CAUTION: - Only values for keys/values in %head_hash_new will be computed and      #
		#                 written in output file!                                                 #
		#               - Keys for listprice and sellprice must exist in %output_hash! No value   #
		#                 is required (will be calculated in this procedure)!                     #
		#                 If one of these keys in not defined in %output_hash, the procedure will #
		#                 fail, due to "length($output_hash{(grep $head_hash_new{$_}..."          #
		#                                                                                         #
		###########################################################################################
		#
		my %output_hash = %{(shift)};
		my %head_hash_new = %{(shift)};
		my %inputargs_hash = %{(shift)};
		my $out_put_add = shift;
		my $error_output;
		#
		foreach( (sort{$a<=>$b}(keys(%output_hash))) )
			{
				#
				# sanitize data
				#
				$output_hash{$_} = del_spaces($output_hash{$_});
			}
		# setting dezimal sign to "." (point)
		foreach( (sort{$a<=>$b}(keys(%head_hash_new))) )
			{
				#
				if ( $head_hash_new{$_} ~~ /price/ || $head_hash_new{$_} ~~ /lastcost/ )
					{
						if ( defined($output_hash{$_}) )
							{
								# check, if "." is already there, like: 1.000,00 => do not change "." to ","!
								if ( grep(/\,/,$output_hash{$_}) )
									{
										# delete points (only decimal point should be left) 
										# => needed for calculation of prices
										$output_hash{$_} =~ s/\.//g;
										# komma to point
										$output_hash{$_} =~ s/\,/\./g;
									}
							}
					}
			}
		#
		if ( ( $inputargs_hash{'clone_price'} ne 0 ) && ( length($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) lt 1 || length($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) lt 1 ) )
			{
				# insert/ clone price
				if ( defined((grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]) )
					{
						if ( defined($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]} = $output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]};
							}
					}
				elsif ( defined((grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]) )
					{
						if ( defined($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]} = $output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]};
							}
					}
			}
		else
			{
				# if both prices given by default
				$inputargs_hash{'clone_price'} = 0;
			}
		#
		# add "make_X", if needed
		#
		if ( defined($inputargs_hash{'make_X'}) && ( (grep $head_hash_new{$_} eq 'make_'.$inputargs_hash{'make_X'} , keys %head_hash_new)[0] ne $inputargs_hash{'make_X'} ) )
			{
				#insert make_X value
				$output_hash{ (grep $head_hash_new{$_} eq 'make_'.$inputargs_hash{'make_X'} , keys %head_hash_new)[0] } = $inputargs_hash{'make_X'};
			}
		#
		# add "model_X", if needed
		#
		if ( defined($inputargs_hash{'model_X'}) && ( (grep $head_hash_new{$_} eq 'model_'.$inputargs_hash{'model_X'} , keys %head_hash_new)[0] ne $inputargs_hash{'model_X'} )  )
			{
				# insert model_X value
				$output_hash{ (grep $head_hash_new{$_} eq 'model_'.$inputargs_hash{'model_X'} , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'partnumber' , keys %head_hash_new)[0]};
			}
		#
		# add "lastcost", if needed
		#
		if ( defined($inputargs_hash{'lastcost'}) && length($output_hash{(grep $head_hash_new{$_} eq 'lastcost' , keys %head_hash_new)[0]}) lt 1 )
			{
				# insert lastcost value
				if ( (grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0] gt 0 )
					{
						if ( defined($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost' , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}*(1-$output_hash{(grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0]}/100);
							}
						elsif(  defined($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost' , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}*(1-$output_hash{(grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0]}/100);
							}
					}
				else
					{
						if ( length($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) gt 0 )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost' , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]};
							}
						elsif(  length($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) gt 0 )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost' , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]};
							}
					}
			}
		#
		# add "lastcost_X", if needed
		#
		if ( defined($inputargs_hash{'lastcost_X'}) && defined($output_hash{(grep $head_hash_new{$_} eq 'lastcost_'.$inputargs_hash{'lastcost_X'} , keys %head_hash_new)[0]}) )
			{
				# insert lastcost_X value
				if ( (grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0] gt 0 )
					{
						if ( defined($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost_'.$inputargs_hash{'lastcost_X'} , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}*(1-$output_hash{(grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0]}/100);
							}
						elsif(  defined($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost_'.$inputargs_hash{'lastcost_X'} , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}*(1-$output_hash{(grep $head_hash_new{$_} eq 'rabattcode' , keys %head_hash_new)[0]}/100);
							}
					}
				else
					{
						if ( defined($output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost_'.$inputargs_hash{'lastcost_X'} , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'listprice' , keys %head_hash_new)[0]};
							}
						elsif(  defined($output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]}) )
							{
								$output_hash{ (grep $head_hash_new{$_} eq 'lastcost_'.$inputargs_hash{'lastcost_X'} , keys %head_hash_new)[0] } = $output_hash{(grep $head_hash_new{$_} eq 'sellprice' , keys %head_hash_new)[0]};
							}
					}
			}
		#
		# add "partnumber prefix", if needed
		#
		if ( defined($inputargs_hash{'pnprefix_add'}) )
			{
				$output_hash{ (grep $head_hash_new{$_} eq 'partnumber' , keys %head_hash_new)[0] } = $inputargs_hash{'pnprefix_add'}."".$output_hash{ (grep $head_hash_new{$_} eq 'partnumber' , keys %head_hash_new)[0] } ;
			}
		#
		# add "description prefix", if needed
		#
		if ( defined($inputargs_hash{'pnprefix_add'}) )
			{
				$output_hash{ (grep $head_hash_new{$_} eq 'description' , keys %head_hash_new)[0] } = $inputargs_hash{'desprefix_add'}." ".$output_hash{ (grep $head_hash_new{$_} eq 'description' , keys %head_hash_new)[0] } ;
			}
		#
		# correcting price factor/ substituion
		#
		if ( defined($output_hash{ (grep $head_hash_new{$_} eq 'price_factor' , keys %head_hash_new)[0] }) )
			{
				if ( length($output_hash{ (grep $head_hash_new{$_} eq 'price_factor' , keys %head_hash_new)[0] }) gt 0 )
					{
						$output_hash{ (grep $head_hash_new{$_} eq 'price_factor' , keys %head_hash_new)[0] } = "pro ".$output_hash{ (grep $head_hash_new{$_} eq 'price_factor' , keys %head_hash_new)[0] };
					}
				else
					{
						$output_hash{ (grep $head_hash_new{$_} eq 'price_factor' , keys %head_hash_new)[0] } = "pro 1";
					}
			}
		#
		# translating unit's into Kivitendo compatible units
		#
		if ( defined((grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0]) )
			{
				if ( defined($unit_trans{$output_hash{ (grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0] }}) )
					{
						# set correct unit value
						$output_hash{ (grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0] } = $unit_trans{$output_hash{ (grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0] }}
					}
				else
					{
						# default value of Kivitendo csv- input is used
						$output_hash{ (grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0] } = 'ERROR_UNIT:'.$output_hash{ (grep $head_hash_new{$_} eq 'unit' , keys %head_hash_new)[0] } ;
					}
			}
		#
		# setting dezimal sign to "," (Komma)
		foreach( (sort{$a<=>$b}(keys(%head_hash_new))) )
			{
				if ( $head_hash_new{$_} ~~ /price/ || $head_hash_new{$_} ~~ /lastcost/ )
					{
						if ( defined( $output_hash{ $_ } ) )
							{
								# check, if "," is already there, like: 1.000,00 => do not change "." to ","!
								if ( !grep(/\,/,$output_hash{$_}) )
									{
										# point to komma
										$output_hash{$_} =~ s/\./\,/g;
									}
							}
					}
			}
		#
		# check for errors
		#
		if ( grep lc($output_hash{$_}) ~~ /error_/ , keys %head_hash_new )
			{
				#
				# store output in error variable
				#
				foreach( (sort{$a<=>$b}(keys(%head_hash_new))) )
					{
						if ( $head_hash_new{$_} ne 'rabattcode' )
							{
								#print "error:".$_."\n";
								$error_output .= $output_hash{$_};
								if( $_ ne (sort{$a<=>$b}(keys(%head_hash_new)))[-1] )
									{
										# if this is not the last value, print row separator sign in file
										$error_output .= $inputargs_hash{'separator'};
									}
							}
					}
			}
		else
			{
				#
				# write output to file (one line)
				#
				foreach( (sort{$a<=>$b}(keys(%head_hash_new))) )
					{
						if ( $head_hash_new{$_} ne 'rabattcode' )
							{
								print $out_put_add $output_hash{$_};
								if( $_ ne (sort{$a<=>$b}(keys(%head_hash_new)))[-1] )
									{
										# if this is not the last value, print row separator sign in file
										print $out_put_add $inputargs_hash{'separator'};
									}
							}
					}
					
				print $out_put_add "\n";
			}
		return $error_output;
	}
#
#############################################################################################
#                                                                                           #
#               delete whitespaces                                                          #
#                                                                                           #
#############################################################################################
#
sub del_spaces
	{
		my($text) = shift @_;
		if ( defined($text) )
			{
				# remove leading whitespace
				$text =~ s/^\s+//;
				# remove trailing whitespace
				$text =~ s/\s+$//;
				# remove unwanted signs (to avoid problems with SQL- statements)
				$text =~ s/\'|\"//g;
				$text =~ s/\`//g;
				#
				# remove protected whitespace
				$text =~ s/ /\ /g;
				# remove multiple whitespaces
				$text =~ s/\ \ */\ /g;
				
			}
		return $text;
	}
#
#############################################################################################
#                                                                                           #
#                create new output file                                                     #
#                                                                                           #
#############################################################################################
#
sub newfile
	{
		my %head_hash_new = %{(shift)};
		my %inputarg_hash_new = %{(shift)};
		my $outputadd = shift;
		#
		# local variables
		#
		my $tmp_outvar=0;
		my $output_file_new=$inputarg_hash_new{'output_file'};
		#
		# "count files"
		#
		if($inputarg_hash_new{'output_filenum'} ne 0)
			{
				print "\n";
				$output_file_new="";
				my @tmp_array=split(/\./,$inputarg_hash_new{'output_file'});
				#
				# combine new file name
				#
				foreach(@tmp_array)
					{
						$output_file_new.=$_;
						if ( $tmp_outvar eq ($#tmp_array-1) )
							{
								$output_file_new.="_".$inputarg_hash_new{'output_filenum'}.".";
							}
						elsif( $tmp_outvar ne $#tmp_array )
							{
								$output_file_new.=".";
							}
						$tmp_outvar++;
					}
				$tmp_outvar=0;
			}
		#
		if ($inputarg_hash_new{'output_filenum'} gt 0)
			{
				close($outputadd);
			}
		print "#                                                    #\n";
		print "#    Lege Ausgabedatei an.....                       #\n";
		print "#                   ",$output_file_new,"\n";
		open(my $output,'>',$output_file_new);
		close($output);
		open($outputadd,'>>',$output_file_new);
		binmode($outputadd, ":utf8");
		print "#                                                    #\n";
		print "#    Schreibe Spaltennamen in erste Zeile            #\n";
		print "#    der neuen Datei.....                            #\n";
		print "#                                                    #\n";
		print "#   | ";
		#
		foreach( (sort{$a<=>$b}(keys(%head_hash_new))) )
			{
				if ( $head_hash_new{$_} ne 'rabattcode' )
					{
						print $head_hash_new{$_};
						print $outputadd $head_hash_new{$_};
						#
						if( $_ ne (sort{$a<=>$b}(keys(%head_hash_new)))[-1] )
							{
								print " | ";
								print $outputadd $inputarg_hash_new{'separator'};
							}
						else
							{
								print " |\n";
								print "#                                                    #\n";
							}
					}
			}
			print $outputadd "\n";
		return $outputadd;
	}
#
#############################################################################################
#                                                                                           #
#                create new output file                                                     #
#                                                                                           #
#############################################################################################
#
sub error_out
	{
		#
		# local variables
		#
		my %error_output_hash = %{shift()};
		my $output_file_new= ${(shift)};
		$output_file_new .='_error';
		# 
		if( scalar(keys %error_output_hash) gt 1 )
			{
				print "#                                                    #\n";
				print "#    ACHTUNG: Es wurden fehlerhafte Datensätze       #\n";
				print "#             gefunden!                              #\n";
				print "#                                                    #\n";
				print "#    Lege Ausgabedatei für fehlerhafte Daten an...   #\n";
				print "#                   ",$output_file_new,"\n";
				open(my $output_error_file,'>',$output_file_new);
				open($output_error_file,'>>',$output_file_new);
				binmode($output_error_file, ":utf8");
				#
				foreach( sort{$a<=>$b}(keys %error_output_hash) )
					{
						print $output_error_file $error_output_hash{$_}."\n";
					}
				close($output_error_file);
			}
		else
			{
				print "#                                                    #\n";
				print "#    Es wurden keine fehlerhaften Datensätze         #\n";
				print "#    gefunden und keine Fehlerdatei angelegt!        #\n";
			}
		
	}
#
#############################################################################################
#                                                                                           #
#                help text (column names)                                                   #
#                                                                                           #
#############################################################################################
#
sub list_tablearg
	{
		my %tableargs_hash = %{(shift)};
		my %tabletrans_hash = %{(shift)};
		#
		print "#  folgende Spaltenbeschriftungen sind verwendbar:\n";
		print "#  Nummer | Spaltenname         | Bedeutung\n";
		print "# ---------------------------------------------\n";
		#
		foreach ( sort{$a<=>$b}(keys(%tableargs_hash)) )
			{
				print "# ";
				my $tmp_var=0;
				#
				while( $tmp_var < (5-length($_)) )
					{
						print " ";
						$tmp_var++;
					}
				print $_,"  |";
				$tmp_var=0;
				#
				while( $tmp_var < (20-length($tableargs_hash{$_})) )
					{
						print " ";
						$tmp_var++;
					}
				print $tableargs_hash{$_}," | ";
				print $tabletrans_hash{$_},"\n";
			}
	}
#
#############################################################################################
#                                                                                           #
#                help text (supplier names)                                                 #
#                                                                                           #
#############################################################################################
#
sub list_suppliers
	{
		my %suppliers_list = %{(shift)};
		my %suppabbr_list = %{(shift)};
		print "#                                                    #\n";
		print "#   Verfügbare Lieferanten:                          #\n";
		print "#                                                    #\n";
		print "#   laufende |                                       #\n";
		print "#     Nummer | Lieferant          | Abkürzung        #\n";
		print "#  --------------------------------------------      #\n";
		#
		foreach ( sort{$a<=>$b}(keys(%suppliers_list)) )
			{
				my $tmp_var=0;
				#
				print "#";
				#
				while( $tmp_var < (10-length($_)) )
					{
						print " ";
						$tmp_var++;
					}
				print $_,"  |";
				#
				$tmp_var=0;
				print " ",$suppliers_list{$_};
				while( $tmp_var < (19-length($suppliers_list{$_})) )
					{
						print " ";
						$tmp_var++;
					}
				print "| ";
				print $suppabbr_list{$_};
				#
				$tmp_var=0;
				while( $tmp_var < (17-length($suppabbr_list{$_})) )
					{
						print " ";
						$tmp_var++;
					}
				print "#\n";
			}
		print "#                                                    #\n";
	}
#
#############################################################################################
#                                                                                           #
#                Programmende                                                               #
#                                                                                           #
#############################################################################################
